--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: FILE
-- @author    Valentin Schmidt
-- @version   0.9
--****************************************************************************

property $

----------------------------------------
-- @constructor
----------------------------------------
on new (me)
  return me
end

--------------------------------------
-- Opens file with specified mode, returns file pointer (xtra instance)
-- @param {string} tFile - path to file
-- @param {string} tMode - r, rb, w, wb, a, ab, rw, rwb, wr, wrb
-- @return {instance|false}
--------------------------------------
on fopen (me, tFile, tMode)
  fp = xtra("fileIO").new()

  case (tMode) of
    "r","rb": fp.openFile(tFile, 1)
    "w","wb":
      --if not fileExists(tFile) then

      fp.openFile(tFile, 2)
      err = fp.status()
      if (err=0) then fp.delete()
      --else if (err and not (err = -37)) then return false

      fp.createFile(tFile)
      fp.openFile(tFile, 2)
    "rw","wr","rwb","wrb": fp.openFile(tFile, 0)
    "a","ab":
      fp.openFile(tFile, 2)
      fp.setPosition(fp.getLength())
    otherwise:
      return 0
  end case

  if (fp.status()<>0) then fp = 0
  return fp
end

--------------------------------------
-- Closes open file
-- @param {instance} fp
--------------------------------------
on fclose (me, fp)
  fp.closeFile()
  fp = 0
end

----------------------------------------
-- Deletes open file
-- @param {instance} fp
-- @return {bool}
----------------------------------------
on fdelete (me, fp)
  err = fp.delete()
  return (err=0)
end

----------------------------------------
-- Reads n bytes from open file, returns bytearray
-- @param {instance} fp
-- @param {integer} n
-- @return {bytearray}
----------------------------------------
on freadbytes(me, fp, n)
  ba = fp.readByteArray(n)
  if ilk(ba)=#bytearray then ba.position = 1
  return ba
end

----------------------------------------
-- Writes byteArray to open file
-- @param {instance} fp
-- @param {bytearray} ba
----------------------------------------
on fwritebytes (me, fp, ba)
  return fp.writeByteArray(ba)
end

----------------------------------------
-- Reads n bytes from open file, returns string
-- @param {instance} fp
-- @param {integer} n
-- @return {string}
----------------------------------------
on freadstring(me, fp, n)
  ba = me.freadbytes(fp, n)
  ba.position = 1
  return ba.readRawString(ba.length)
end

----------------------------------------
-- Writes string to open file
-- @param {instance} fp
-- @param {string} str
----------------------------------------
on fwritestring (me, fp, str)
  fp.writeString(str)
end

--------------------------------------
-- Returns position in open file
-- @param {instance} fp
-- @return {integer}
--------------------------------------
on ftell (me, fp)
  return fp.getPosition()
end

--------------------------------------
-- Sets position in open file
-- whence: either integer or symbol
-- 0 or #SEEK_SET: set position to offset bytes. (default)
-- 1 or #SEEK_CUR: set position to current position plus offset.
-- 2 or #SEEK_END: set position to end of file plus offset.
-- @param {instance} fp
-- @param {integer} pos
-- @param {integer|symbol} [whence=0]
-- @return {bool}
--------------------------------------
on fseek (me, fp, pos, whence)
  if voidP(whence) then whence = 0
  else if ilk(whence)=#symbol then
    whence = [#SEEK_SET, #SEEK_CUR, #SEEK_END].getPos(whence)
    if whence>0 then whence=whence-1
  end if

  case (whence) of
    0: fp.setPosition(pos)
    1: fp.setPosition(fp.getPosition() + pos)
    2: fp.setPosition(fp.getLength() + pos)
  end case

  return (fp.status()=0)
end

--------------------------------------
-- Returns size of open file
-- @param {instance} fp
-- @return {integer}
--------------------------------------
on fsize (me, fp)
  return fp.getLength()
end

----------------------------------------
-- Reads 32-bit integer from open file
-- @param {instance} fp
-- @param {symbol} [endian=#LittleEndian]
-- @return {integer}
----------------------------------------
on freadint32 (me, fp, endian)
  ba = fp.readByteArray(4)
  ba.position = 1
  if not voidP(endian) then ba.endian = endian
  return ba.readInt32()
end

----------------------------------------
-- Writes 32-bit integer to open file
-- @param {instance} fp
-- @param {integer} n
-- @param {symbol} [endian=#LittleEndian]
----------------------------------------
on fwriteint32 (me, fp, n, endian)
  ba = bytearray(4)
  if not voidP(endian) then ba.endian = endian
  ba.writeInt32(n)
  fp.writeByteArray(ba)
end

----------------------------------------
-- Reads 16-bit integer from open file
-- @param {instance} fp
-- @return {integer}
----------------------------------------
on freadint16 (me, fp)
  ba = fp.readByteArray(2)
  ba.position = 1
  return ba.readInt16()
end

----------------------------------------
-- Writes 16-bit integer to open file
-- @param {instance} fp
-- @param {integer} n
----------------------------------------
on fwriteint16 (me, fp, n)
  ba = bytearray(2)
  ba.writeInt16(n)
  fp.writeByteArray(ba)
end

----------------------------------------
-- Reads 8-bit integer from open file
-- @param {instance} fp
-- @return {integer}
----------------------------------------
on freadint8 (me, fp)
  ba = fp.readByteArray(1)
  return ba[1]
end

----------------------------------------
-- Writes 8-bit integer to open file
-- @param {instance} fp
-- @param {integer} n
----------------------------------------
on fwriteint8 (me, fp, n)
  fp.writeByteArray(bytearray(1, n))
end

----------------------------------------
-- Copies data from fp_src to fp_dest, using a memory friendly buffer
-- @param {instance} fp_src
-- @param {instance} fp_dest
-- @param {integer} size
-- @param {integer} [buffsize=10240] - optional, default = 10 KB buffer
----------------------------------------
on fcopybytes (me, fp_src, fp_dest, size, buffsize)
  if voidP(buffsize) then buffsize = 10*1024 -- 10 KB
  cnt = size/buffsize
  repeat with i = 1 to cnt
    fp_dest.writeByteArray(fp_src.readByteArray(buffsize))
  end repeat
  if size>cnt*buffsize then
    fp_dest.writeByteArray(fp_src.readByteArray(size-cnt*buffsize))
  end if
end

--**************************************
-- global functions, no file pointer passed
--**************************************

----------------------------------------
-- Returns file size
-- @param {string} tFile
-- @return {integer}
----------------------------------------
on size (me, tFile)
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 1)
  err = fp.status()
  if err then return err
  len = fp.getLength()
  fp.closeFile()
  fp = 0
  return len
end

----------------------------------------
-- Truncates file
-- @param {string} tFile
-- @param {integer} tSize
-- @return {bool}
----------------------------------------
on truncate (me, tFile, tSize)
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 0)
  data = fp.readByteArray(tSize)
  fp.delete()
  fp.createFile(tFile)
  fp.openFile(tFile, 2)
  fp.writeByteArray(data)
  err = fp.status()
  fp.closeFile()
  fp=0
  return (err=0)
end

----------------------------------------
-- Deletes file
-- @param {string} tFile
-- @return {bool}
----------------------------------------
on delete (me, tFile)
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 0)
  err = fp.status()
  if not (err) then err = fp.delete()
  fp = 0
  return (err=0)
end

----------------------------------------
-- Checks if file exists
-- @param {string} tFile
-- @return {bool}
----------------------------------------
on exists (me, tFile)
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 1)
  stat = fp.status()
  fp.closeFile()
  fp = 0
  -- 0: OK: item is a file
  -- -36: I/O Error: item is a folder
  -- -37: Bad File Name: does not exist
  return (stat=0)
end

----------------------------------------
-- Reads whole file, returns string
-- @param {string} tFile
-- @param {string} [tEncoding="utf-8"]
-- @return {string|false}
----------------------------------------
on getString (me, tFile, tEncoding)
  if voidP(tEncoding) then tEncoding="utf-8"
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 1)
  err = fp.status()
  if (err) then return false
  if tEncoding<>"utf-8" then fp.setCharSet(tEncoding)
  ret = fp.readFile()
  fp.closeFile()
  fp = 0
  if chartonum(ret.char[1]) = 65279 then
    delete char 1 of ret -- remove UTF-8 BOM
  end if
  return ret
end

----------------------------------------
-- Saves string as file
-- @param {string} tFile
-- @param {string} tString
-- @param {bool} [tAddBOM=false]
-- @param {string} [tEncoding="utf-8"]
-- @return {bool} success
----------------------------------------
on putString (me, tFile, tString, tAddBOM,  tEncoding)
	if voidP(tEncoding) then tEncoding="utf-8"
	fp = xtra("fileIO").new()
	fp.openFile(tFile, 2)
	err = fp.status()
	if not (err) then fp.delete()
	else if (err and not (err = -37)) then return false
	fp.createFile(tFile)
	err = fp.status()
	if (err) then return false
	fp.openFile(tFile, 2)
	err = fp.status()
	if (err) then return false
	if tEncoding<>"utf-8" then fp.setCharSet(tEncoding)
	if tEncoding="utf-8" AND tAddBOM then
	  fp.writeString(numtochar(65279)) -- add UTF-8 BOM
	end if
	fp.writeString(tString)
	fp.closeFile()
	fp=0
	return true
end

----------------------------------------
-- Returns file as ByteArray
-- @param {string} tFile
-- @return {byteArray|false}
----------------------------------------
on getBytes (me, tFile)
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 1)
  err = fp.status()
  if (err) then return false
  len = fp.getLength()
  if len then
  	data = fp.readByteArray(len)
  else
  	data = bytearray()
  end if
  fp.closeFile()
  fp = 0
  return data
end

----------------------------------------
-- Saves ByteArray to file
-- @param {string} tFile
-- @param {byteArray} tString
-- @param {integer} [tLength=VOID]
-- @return {bool}
----------------------------------------
on putBytes (me, tFile, tByteArray, tLength)
  fp = xtra("fileIO").new()
  fp.openFile(tFile, 2)
  err = fp.status()
  if not (err) then fp.delete()
  else if (err and not (err = -37)) then return false
  fp.createFile(tFile)
  err = fp.status()
  if (err) then return false
  fp.openFile(tFile, 2)
  err = fp.status()
  if (err) then return false
  if voidP(tLength) then tLength = tByteArray.length
  fp.writeByteArray(tByteArray, 1, tLength)
  fp.closeFile()
  fp=0
  return true
end

----------------------------------------
-- Displays open file dialog
-- @param {string} tTitle
-- @param {string} filterMask
-- @param {string} defaultFileName
-- @return {string}
----------------------------------------
on selectFileOpen (me, tTitle, filterMask, defaultFileName)
  fio = xtra("fileIO").new()
  fio.setFilterMask(filterMask)
  fn = fio.displayOpen()
  fio = 0
  return string(fn)
end

----------------------------------------
-- Displays save file dialog
-- @param {string} tTitle
-- @param {string} filterMask
-- @param {string} defaultFileName
-- @return {string}
----------------------------------------
on selectFileSave (me, tTitle, filterMask, defaultFileName)
  fio = xtra("fileIO").new()
  fio.setFilterMask(filterMask)
  fn = fio.displaySave(tTitle, defaultFileName)
  fio = 0
  return string(fn)
end

----------------------------------------
-- Select folder via save file dialog
-- @param {string} tTitle
-- @return {string}
----------------------------------------
on selectFolder (me, tTitle)
  if voidP(tTitle) then tTitle = "Select Folder"
  fio = new(xtra "fileIO")
  fn = fio.displaySave(tTitle, "*.*")
  err = fio.status()
  fio = 0
  if string(fn)=EMPTY then return ""
  if err<>0 then
    put "Save File Error:" && fio.error(err)
    --exit
    return ""
  end if
  pd = the last char of _movie.path
  od = _player.itemDelimiter
  _player.itemDelimiter = pd
  delete the last item of fn
  _player.itemDelimiter = od
  put pd after fn
  return fn
end

----------------------------------------
-- Returns file type
-- @param {string} tFilename
-- @return {string}
----------------------------------------
on getFileType (me, tFilename)
  od = the itemdelimiter
  the itemdelimiter = "."
  tType = the last item of tFilename
  the itemdelimiter = od
  
  -- to lower
  alphabet = "abcdefghijklmnopqrstuvwxyz" --äöü"
  cnt = tType.length
  repeat with i = 1 to cnt
    pos = offset(tType.char[i],alphabet)
    if pos > 0 then put alphabet.char[pos] into char i of tType
  end repeat
  
  return tType
end

----------------------------------------
-- Extracts filename from path
-- @param {string} tPath
-- @return {string}
----------------------------------------
on getFileName (me, tPath)
  od = the itemdelimiter
  the itemdelimiter = $.PD
  tFileName = the last item of tPath
  the itemdelimiter = od
  return tFileName
end

----------------------------------------
-- Extracts path to directory from path
-- @param {string} tPath
-- @return {string}
----------------------------------------
on getFilePath (me, tPath)
  od = the itemdelimiter
  the itemdelimiter = $.PD
  delete the last item of tPath
  the itemdelimiter = od
  return tPath
end

----------------------------------------
-- Extracts basename from path
-- @param {string} tPath
-- @return {string}
----------------------------------------
on getBaseName (me, tPath)
	fn = me.getFileName(tPath)
	cnt = fn.length
	repeat with i = cnt down to 1
		if fn.char[i]="." then return fn.char[1..i-1]
	end repeat
  return fn
end
